/****************************************************************************
 *                                                                          *
 * This program called cvi (C verbal interpreter) looks for information     *
 * about variables and data types in an ANSI C program.                     *
 * Copyright (c) 1999 Chepelianskii Alexeii.                                *
 *                                                                          *
 * This program is free software; you can redistribute it and/or modify     *
 * it under the terms of the GNU General Public License as Published by     *
 * the Free Software Fondation; either version 2 of the License, or (at     *
 * your option) any later version.                                          *
 *                                                                          *
 * This program is distributed in the hope it will be useful, but WITHOUT   *
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    *
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License     *
 * for more details.                                                        *
 *                                                                          *
 * You should have received a copy of the GNU General Public License along  *
 * with this program; if not, write to the Free Software Foundation, Inc.,  *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                  *
 *                                                                          *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "trad.h"

static void lirecond(Condition *, int argc, char **argv);
static void anafile(Variable **, Variable **, int *);
static void removetree(Variable *p);
static void disp(Condition *, Variable *);
static void setInput(FILE *);
static Variable *realremove(Variable *, Variable **, int);

int nlignes = 1;
int nblocs = 0;
char *filename = INPUT;

static Condition cond = {0};
static const char *genericName = GENER;


int main(int argc, char **argv)
{
    Variable *root = NULL;
    Variable *result = NULL;
    char **files;
    FILE *input;
    int correct = 1;

    /* Command line arguments parsing. */
    lirecond(&cond, argc, argv);
    /* Input files seting. */
    if (cond.files != NULL) {
       for (setToFirstG(&cond.files); isElemG(cond.files); 
	    setToNextG(&cond.files)) {
	  files = (char **) getG(cond.files);
	  if ((input = fopen(*files, "r")) == NULL) {
	     fprintf(stderr,"error : file %s does not exist\n", *files);
	     correct = 0;
	     continue;
	  }
	  setInput(input);
	  filename = *files;
	  anafile(&root, &result, &correct);
	  fclose(input);
	  nlignes = 1;
       }
    }
    else {
       setInput(stdin);
       anafile(&root, &result, &correct);
    }
    if (!correct || iserror())
       return 1;
    /* Clean the list of variables and print the collected information. */
    realremove(root, &result, 1);
    disp(&cond, result);
    return 0;
}

/* Each command line argument is defined by a constant to 
   make it more easy to change option names. */
static const char *name = "-n"     , *nameEnd = "-N"   , 
                  *bloc = "-b"     , *blocEnd = "-B"   ,
                  *str = "-s"      , *strEnd = "-S"    ,
                  *enumt = "-e"    , *enumEnd = "-E"   , 
                  *type = "-t"     , *typeEnd = "-T"   ,
                  *membersOf = "-m", *membersEnd = "-M",
                  *file = "-f"     , *fileEnd = "-F"   ,
                  *macro = "-p"    , *macroEnd = "-P"  ,
                  *min = "-min"    , *max = "-max"     ,
                  *help = "--help";

static Variable *StrictAjouter(Variable *p, const struct infovar *varl, int n);

/* A fonction to parse the -b -n options when they are used in a bloc. */
static int SubBloc(char **func, char ***argv, int *argc)
{
    if (!strcmp(**argv, bloc)) {
       if (!--*argc)
	  fatalerror("bad bloc");
       *func = dupchaine(*++(*argv));
       return 1;
    } else if (!strcmp(**argv, name)) {
       *func = (char *) genericName;
       return 1;
    }
    return 0;
}

static void BasicBloc(const char *begin, const char *End, 
		      char ***argv, int *argc, Variable **var)
{
    struct infovar inf = {0};
    char *CurrFunc = (char *) genericName;
    for ((*argv)++, (*argc)-- ; ; (*argv)++, (*argc)--) {
       if (*argc <= 0)
	  fatalerror("usage, %s -b bloc var_of_bloc1... -b ... -n ... %s", 
		begin ,End);
       if (!strcmp(**argv, End))
	  break;
       if (SubBloc(&CurrFunc, argv, argc)) 
	  ;
       else {
	  inf.nom = dupchaine(**argv);
	  inf.fonction = CurrFunc;
	  *var = StrictAjouter(*var, &inf, 1);
       }
    }
}

static void printHelp(void);

/* Parse conditions defined buy the command line. */
static void lirecond(Condition *cond, int argc, char **argv)
{
    const struct infovar ToClean = {0};
    struct infovar inf = ToClean;

    for (++argv, --argc ; argc > 0; argv++, --argc, inf = ToClean) {
       if (!strcmp(*argv, help)) {
	  printHelp();
       } else if (!strcmp(*argv, min)) {
	  cond->affmin = 1;
       } else if (!strcmp(*argv, max)) {
	  cond->affmax = 1;
       } else if (!strcmp(*argv, str)) {
	  cond->affstruct = 1;
	  BasicBloc(str, strEnd, &argv, &argc, &cond->Affstruct);
       } else if (!strcmp(*argv, type)) {
	  cond->afftypedef = 1;
	  BasicBloc(type, typeEnd, &argv, &argc, &cond->Afftypedef);
       } else if (!strcmp(*argv, enumt)) {
	  cond->affenum = 1;
	  BasicBloc(enumt, enumEnd, &argv, &argc, &cond->Affenum);
       } else if (!strcmp(*argv, macro)) {
	  cond->affmacro = 1;
	  BasicBloc(macro, macroEnd, &argv, &argc, &cond->Affmacro);
       } else if (!strcmp(*argv, name)) {
	  BasicBloc(name, nameEnd, &argv, &argc, &cond->var);
       } else if (!strcmp(*argv, bloc)) {
	  inf.nom = (char *) genericName;
	  for (argv++, --argc ; ; argv++, --argc) {
	     if (argc <= 0)
	        fatalerror("usage, -b list_of_functions -B\n");
	     else if (!strcmp(*argv, blocEnd))
	        break;
	     inf.fonction = dupchaine(*argv);
	     cond->bloc = StrictAjouter(cond->bloc, &inf, 1);
	  }
       } else if (!strcmp(*argv, file)) {
	  char **mem;
	  for (argv++, --argc ; ; argv++, --argc) {
	     if (argc <= 0)
	        fatalerror("usage, -f list_of_files_to_treat -F\n");
	     if (!strcmp(*argv, fileEnd))
	        break;
	     mem = (char **) pushG(sizeof(char *), &cond->files);
	     *mem = dupchaine(*argv);
	  }
       } else if (!strcmp(*argv, membersOf)) {
	  char *nom = NULL;
	  int size, i, step;
	  char *func = (char *) genericName;
	  struct infovar inf = {0};

	  for (argv++, --argc ; ; argv++ , --argc) {
	     if (!argc)
	        fatalerror("usage:membersOf -bloc -name structure/member -fO");
	     if (!strcmp(*argv, membersEnd))
	        break;
	     if (SubBloc(&func, &argv, &argc))
	        ;
	     else {   
	        for (step =40 ,i =0 ,size =0 ,nom =NULL ; **argv ; (*argv)++) {
		   /* 
		      To cause explicit warnings when $nom is not correctly
		      set, I analyze *argv letter by letter.
		   */
		   if (isalnum(**argv) || **argv == '/' || **argv == SPECIAL) {
		      if (i + 1 >= size && (nom = (char *) realloc((void *) 
			     nom, (size += step) * sizeof(char))) == NULL)
			    AllocError();
		      nom[i++] = **argv;
		   } else
		      fatalerror("syntax : name_of_structure/name_of_members");
		}
		nom[i] = '\0';
		inf.nom = nom;
		inf.fonction = func;
		cond->str = StrictAjouter(cond->str, &inf, 1);
	     }
	  }
       } else
	 for (; **argv; (*argv)++)
	  switch (*(*argv)) { /* get the next letter */
	  case 'e':
	      cond->affenum = !cond->affenum;
	      break;
	  case 't':
	      cond->afftypedef = !cond->afftypedef;
	      break;
	  case 's':
	      cond->affstruct = !cond->affstruct;
	      break;
	  case 'p':
	      cond->affmacro = !cond->affmacro;
	      break;
	  case 'n':
	      cond->noaffvar = !cond->noaffvar;
	      break;
	  case 'f':
	      if (cond->files != NULL)
		 cleanG(&cond->files);
	      cond->files = NULL;
	      break;
	  default :
	      fatalerror("unknown option %s, type cvi %s for help",
			 *argv, help);
	      break;
	  }
    }
}

static setToZero(Variable **v)
{
    if (*v != NULL) {
        setToZero(&(*v)->gauche);
        (*v)->ntrouves = 0;
	setToZero(&(*v)->droit);
    }
}

static void storeCopy(Variable *v, char *name, char *func);
static void affDefCopy(void);

static void lireRes(char **word, int *SizeOfWord, char *func, 
		    Variable **ListOfVariables)
{
    register int trv;

    do {
       trv = 0;
       while (**word == '[')
	  if (anaind(ListOfVariables, func, word, SizeOfWord) != ']')
	     remettrecar(0); /* cause a syntax error */
	  else
	     liremot(word, SizeOfWord);
       while (**word == ')' || **word == '(' || **word == '*') {
	  liremot(word, SizeOfWord);
	  trv = 1;
       }
    } while (trv);
}


/* Identify a Variable from its name and bloc */
int unknownvar(Variable **racine, char *currentFunc, char **mot, int *size)
{
    Variable *trouve = *racine;
    Variable *copy = NULL;
    Variable *tmp;
    char *name = NULL;
    int res = 0;

    for ( ; ; ) {
       if ((trouve = trouverDansArbre(trouve, *mot, currentFunc)) == NULL)
	  break;
       trouve->ntrouves++;
       res = 1;
       if (copy != NULL && copy->copyOfStr != NULL && 
	                              !strcmp(name, copy->copyName)) {
	  copy = trouverDansArbre(copy->copyOfStr, *mot, currentFunc);
	  copy->ntrouves++;
	  copy = NULL;
       }
       if (name != NULL)
	  mstrcat(&name, "/");
       mstrcat(&name, *mot);
       if ((tmp = StrictTrouver(cond.str, name, currentFunc)) != NULL &&
	                trouve->copyOfStr == NULL ) {
	  if (trouve->var.str != NULL) {
	     copierVar(&trouve->copyOfStr, trouve->var.str->obj.dec);
	     setToZero(&trouve->copyOfStr);
	     storeCopy(trouve->copyOfStr, name, currentFunc);
	     trouve->copyName = dupchaine(name);
	  }
       }
       liremot(mot, size);
       if (trouve->var.str != NULL && trouve->var.str->obj.dec != NULL) {
	  /*  does not care if the code is really correct */
	  lireRes(mot, size, currentFunc, racine /* the main variable list */);
	  if (!strcmp(*mot , ".") || !strcmp(*mot, "->")) {
	     liremot(mot, size);
	     lireRes(mot, size, currentFunc, &trouve);
	     if (trouve->copyOfStr != NULL)
	        copy = trouve;
	     trouve = trouve->var.str->obj.dec;
	     continue;
	  } 
       }
       remettrech(*mot);
       break;
    }
    free((void *) name);
    return res;
}


/* routine to parse a file which contains a C source code */
static void anafile(Variable **racine, Variable **res, int *cor)
{
    char *mot = NULL;
    int size = 0;
    int c, typevar;
    int typeder = 0;        /* type of the last word liremot read.*/
    int dansBlocDerive = 0;
    #define MAXMOT 300
    char currentFunc[MAXMOT] = EXTERN;

    struct infovar quellevar;

    while ((c = liremot(&mot, &size)) != EOF) {
       if (c == '{') {
	  ++nblocs;
	  continue;
       } else if (c == '}') {
	  if (!--nblocs)
	     strcpy(currentFunc, EXTERN);
	  else if (nblocs == 1)
	     dansBlocDerive = 0;
	  else if (nblocs < 0)
	     error("error unexpected '}' line %d", nlignes), *cor = 0;
	  clean();
	  *racine = realremove(*racine, res, 0);
	  continue;
       } else if (c == '\0') 
	  error("syntax, line %d", nlignes), *cor = 0;

       /* to detect "main" as a function, all the others functions
	  are presumed to be prototyped correctly */
       if (!nblocs && !strcmp(mot, "main")) {
	  remettrech(" main");
	  strcpy(mot, "int");
       }

       if (unknownvar(racine, currentFunc, &mot, &size))
	  /* unknownvar identify a variable. */;
       else if (IsType(mot)) {
	  /* new variable detected */
	  remettrech(mot);
          typevar = trad(racine, &quellevar, currentFunc);

	  /* To allow nested functions used in gcc */
	  if (typevar == MENTION)
	     /* save the name of the current function */
	     strcpy(currentFunc, quellevar.nom);
	  else if (typevar == OLD_MENTION) {
	     strcpy(currentFunc, quellevar.nom);
	     nblocs++;
	     /* Wait for the end of the definition. */
	     while ((c = liremot(&mot, &size)) != '{') {
	        remettrech(mot);
		if (trad(racine, &quellevar, currentFunc) != DECLARATION)
		   error("wrong definition of %s line %d", 
                          quellevar.nom, nlignes);
	     }
	  } else if (typevar == CAST && typeder != '(') {
	     error("useless parentheses line %d", nlignes), *cor = 0;
	  } else if (typevar == ERREUR )
	     error("line %d",nlignes), *cor = 0;
       } else if (!findMacro(mot, currentFunc)) {
	  /* An error occured in a macro */
	  *cor = 0;
       }
       typeder = c;
    }
}


/* To delete a variable from pp */
static void removepart(Variable **pp)
{
     Variable *p, *q, **qq;
     if (*pp != NULL) {
        p = *pp;
        if (p->gauche == NULL) {
	   *pp = p->droit;
	   free(p);
	} else if (p->droit == NULL) {
	   *pp = p->gauche;
	   free(p);
	} else {
	   qq = &p->gauche;
	   while ((*qq)->droit != NULL)
	      qq = &(*qq)->droit;
	   q = *qq;
	   *qq = q->gauche;
	   p->var = q->var;
	   p->ntrouves = q->ntrouves;
	   free((void *) q);
	}
     }
}

/* 
 * To move all variables from p, that are not in scope to dest
 * if forced == 1 all variables will be moved.
 */
static Variable *realremove(Variable *p, Variable **dest, int forced)
{
    if (p != NULL) {
       p->gauche = realremove(p->gauche, dest, forced);
       if (p->var.status > nblocs || forced) {
	  if (! p->var.def && !forced)
	     p->var.status = RESERVED;
          else {
	     p->var.status = OLD;
	     *dest = StrictAjouter(*dest, &p->var, p->ntrouves);
	     removepart(&p);
	     return p = realremove(p, dest, forced);
	  }
       }
       p->droit = realremove(p->droit, dest, forced);
    }
    return p;
}


/* removetree : cleans the binnary construction at the end of a bloc */
static void removetree(Variable *p)
{
    if (p != NULL) {
       removetree(p->gauche);
       if (p->var.status > nblocs)
	  if (!p->var.def)
	     p->var.status = RESERVED;
          else
	     p->var.status = OLD;
       removetree(p->droit);
    }
}

/* to hide all the variables of the current bloc. */
void newbloc(Variable *base)
{
    clean();
    removetree(base);
}


/* allouernoeud : allocate space in memory for a new variable. */
inline static Variable *allouernoeud(void)
{
    return (struct variable *) malloc(sizeof(Variable));
}

/* ajoutarbre : adds a variable to the variable list pointed by p. */
Variable *ajoutarbre(Variable *p, struct infovar *varl)
{
    int cond;
    if (p == NULL) {
       /* fill the new structure with the information of varl */
       p = allouernoeud();
       p->var = *varl;
       p->var.fonction = dupchaine(varl->fonction);
       p->var.file = filename;
       p->ntrouves = 1;
       p->copyOfStr = p->gauche = p->droit = NULL;
    } else {
       if (!(cond = strcmp(varl->nom, p->var.nom))) {

	     int bloc = !strcmp(varl->fonction, EXTERN) ? 0 :varl->status;

	     if (p->var.status == bloc || p->var.status == RESERVED && !bloc) {
	        /* to save the definition not the declartion of the variable */

	        if (varl->def && p->var.def || 
		    varl->type != NULL  && p->var.type != NULL && 
		                          strcmp(varl->type, p->var.type) ||
		    (p->var.type == NULL) != (varl->type == NULL) ){

		    if (p->var.strtype != NULL && varl->strtype != NULL &&
			   ((motcmp(p->var.strtype, "static") || 
			   motcmp(varl->strtype, "static")) &&
			   !bloc && !p->var.status)) {
		        /* to assure that the following code will be
			   parsed correctly :
			   file 1 : static A; file 2 : static A */
		        /* to erase the old variable */
		        p->var.status = OLD;
			/* creates the new one */
			p->droit = ajoutarbre(p->droit, varl);
			return p;
		    }
		    error("wrong variable definition %s  line %d",
			                     varl->nom, nlignes);
		}
	        if (!p->var.def) {
		    if (p->var.strtype != NULL) 
		       free((void *) p->var.strtype);
		    p->var.strtype = varl->strtype;
		    p->var.text = varl->text;
		    p->var.ligne = varl->ligne;
		}
		p->ntrouves++;
		if (!p->var.def)
		   p->var.status = bloc;
		if (varl->def)
		   p->var.def = 1;
		return p;
	     } else if (p->var.status < bloc && p->var.status != RESERVED 
			        && p->var.status != OLD)
	        cond = -1;	
     else
	        if (!(cond = strcmp(varl->fonction, p->var.fonction)))
		   cond = 1;
       }
       if (cond < 0)
          p->gauche = ajoutarbre(p->gauche, varl);
       else
          p->droit = ajoutarbre(p->droit, varl);
    }
    return p;
}

/* 
 * trouverDansArbre : returns a pointer to the matching variable defined
 * by its name and bloc, NULL if no one is found. 
 */
Variable* trouverDansArbre(Variable *p, char *m, char *f)
{
    int cond;
    Variable *renc;
    static Variable *ext = NULL;

    if (p == NULL) {
       Variable *temp = ext;
       ext = NULL;
       return temp;
    } else {
       if (!(cond = strcmp(m, p->var.nom))) {
	  if (p->var.status == nblocs) {
	     ext = NULL;
	     return p;
	  } else if (p->var.status < nblocs && p->var.status != RESERVED && 
		     p->var.status != OLD) {
	        cond = -1;
	  } else
	     if (!(cond = strcmp(f, p->var.fonction)))
	        cond = 1;
	  if (ext == NULL && p->var.status != RESERVED && p->var.status != OLD)
	        ext = p;
	  else if (ext != NULL && p->var.status > ext->var.status)
	        ext = p;
       }
       if (cond < 0)
          renc = trouverDansArbre(p->gauche, m, f);
       else
          renc = trouverDansArbre(p->droit, m, f);
    }
    return renc;
} 

/*
 * Equivalent to ajoutarbre execpt the way cond is set. As it is the main
 * part of the stuff in ajoutarbre it was difficult to built a more generic 
 * version.
 */
static Variable *StrictAjouter(Variable *p, const struct infovar *varl, int n)
{
    int cond;
    if (!p) {
       p = allouernoeud();
       p->var = *varl;
       p->copyOfStr = p->gauche = p->droit = NULL;
       p->ntrouves = n;
    } else {
       if (!(cond = strcmp(varl->nom, p->var.nom)))
	  if (!(cond = strcmp(varl->fonction, p->var.fonction)))
	     return p;
       if (cond <= 0)
	  p->gauche = StrictAjouter(p->gauche,(struct infovar *)varl, n);
       else
	  p->droit = StrictAjouter(p->droit,(struct infovar *)varl, n);
    }
    return p;
}


Variable *StrictTrouver(Variable *p, char *m, char *f)
{
    int cond;
    if (!p)
       return NULL;
    if (!(cond = strcmp(m, p->var.nom)) || !strcmp(m, genericName) ||
	!strcmp(p->var.nom, genericName))
       if (!(cond = strcmp(f, p->var.fonction)) || !strcmp(f, genericName) ||
	   !strcmp(p->var.fonction, genericName))
	  return p;
    if (cond < 0)
       return StrictTrouver(p->gauche, m, f);
    else
       return StrictTrouver(p->droit, m, f);
}


/* Helper routine for affarbre */
static void conformCond(const struct infovar *var, const Condition *cond, 
			                                               int *a)
{
    register condFound = 0;
    if (cond->var != NULL) {
       condFound = 1;
       if (StrictTrouver(cond->var,var->nom,var->fonction) != NULL)
	  return;
    } else if (cond->bloc != NULL) {
       condFound = 1;
       if (StrictTrouver(cond->bloc,(char *) genericName, var->fonction) 
	    != NULL)
	  return;
    }
    if (condFound)
       *a = 0;
    return;
}

static void printVar(Variable *p)
{
    char *str1 = NULL;
    vastrcat(2, &str1, p->var.text != NULL ? p->var.text : " ",
	               p->var.strtype);
    if (str1 == NULL)
       AllocError();
    putchar('\n');
    printf("%s is declared as :\n", p->var.nom);
    maketable(1, 5, 0, 65, str1);
    printf("     -> found %d time%s, defined in %s, line %d of %s\n",
           p->ntrouves, p->ntrouves == 1 ? "" : "s",
           p->var.fonction, p->var.ligne, p->var.file);
}

/*
 * printMinMax : as its name stresses just prints out definitions
 * of variables that are very or almost not used by the parsed 
 * program.
 */
static void printMinMax(int v, char *s, list **l)
{
    Variable **tmp;
    if (v) {
       printf("%s used variable%s\n", s, howManyG(*l) == 1 ? "": "s");
       printf("-----------------------\n");
    }
    for (setToFirstG(l); tmp = (Variable **) getG(*l) ;setToNextG(l))
       printVar(*tmp);
    printf("\n");
}

/* 
 * affarbre : prints out the p tree in an ordered way,
 *            and selects the most and less used variables.
 */
void affarbre(Variable *p, const Condition *cond)
{
    register Variable **tmp;
    int afficher = (!cond) ? 1 : !cond->noaffvar;
    static int count = 0;       /* recursive levels counter */
    static list *Min = NULL, *Max = NULL;
    static int min, max;

    count++;
    if (p != NULL) {
       affarbre(p->gauche, cond);
       if (cond != NULL) {
	  /* This is to find out most and less used variables */
	  /* Maximum */
	  if (cond->affmax)
	     if (!Max || p->ntrouves >= max) {
	        if (p->ntrouves > max || !Max) cleanG(&Max), max = p->ntrouves;
	        if (!(tmp = (Variable **) pushG(sizeof(Variable *), &Max)))
		   AllocError();
		*tmp = p;
	     }
	  /* Minimum */
	  if (cond->affmin)
	     if (!Min || p->ntrouves <= min) {
	        if (p->ntrouves < min || !Min) cleanG(&Min), min = p->ntrouves;
		if (!(tmp = (Variable **) pushG(sizeof(Variable *), &Min)))
		   AllocError();
		*tmp = p;
	     }
	  /* condition pasing */
	  conformCond(&p->var, cond, &afficher);
       } 
       if (afficher)
	  printVar(p);
       affarbre(p->droit, cond);
    }
    if (! --count && cond != NULL) {
       if (cond->affmax || cond->affmin)
	  printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *             Most and less used variables.            *\n"
           "\t ********************************************************/\n"
	   "\n");
       printMinMax(cond->affmax, "most", &Max);
       printMinMax(cond->affmin, "less", &Min);
    }  
}


/* affDefMembers : works a like affarbre however it's simpler
                   (doesn't compute maximums, minimums) and produces
		   a different output. */
void affDefMembers(Variable *p)
{
    if (p != NULL) {
       char *str1 = NULL, *str2 = NULL;
       affDefMembers(p->gauche);
       vastrcat(2, &str1, p->var.text != NULL ? p->var.text : " " , 
		          p->var.strtype);
       vastrcat(4 ,&str2, "used ", 
                         conv(p->ntrouves), 
                         " time",
                         p->ntrouves == 1 ? "" : "s");
       maketable(4, 5, 2, 12, p->var.nom , 3, ":", 31, str1, 20, str2);
       affDefMembers(p->droit);
    }
}

/* 
 * disp : prints the information, stored in 'racine', or from other 
 *        information sources, the real output depends of the given 
 *	  Condition structure.
 * It is the principal output routine. 
 */
static void disp(Condition *cond, Variable *racine)
{
    void affDefVar(Variable *, const Condition *);

    if (cond->affstruct)
       affDefStruc(cond);
    affDefCopy();
    if (cond->afftypedef)
       dispType(cond);
    if (cond->affenum)
       affDefEnum(cond);
    if (cond->affmacro)
       affDefMacro(cond);
    if (!cond->noaffvar || cond->affmax || cond->affmin)
       affDefVar(racine, cond);  
}


/* readword : to read the next word from the input. */
int readword(char **mot, int *lim)
{
    int c, i, lirecar(void);
    void remettrecar(int);
    const step = 300;
    char *m;

    if (*mot == NULL)
       *mot = (char *) malloc((*lim += step) * sizeof(char));
    m = *mot;
    i = 0;
    do {
       c = lirewspace();
       if (c != EOF)
	  m[i++] = c;
    } while (c == '#' && i < 2);
    if (!isalpha(c) && c != '_' && c != SPECIAL) {
                            /* for the #define C preprocessor directive */
       if (c == '-')
	  if ((c = lirecar()) == '>') {
	                        /* for the -> operator */
	     m[i++] = c;
	     m[i++] = '\0';
	     return m[0];
          } else
	     remettrecar(c);
       m[i] = '\0';
       if (!negligerComment(c) && !negligerGuillemets(c))
	  return c;
    }
    for ( ; ; i++) {
       if (i + 1 >= *lim) {
	  *mot = (char *) realloc((void *)*mot, (*lim += step)*sizeof(char));
	  m = *mot;
       }
       if (!isalnum(m[i] = lirecar()) && m[i] != '_') {
          remettrecar(m[i]);
          break;
       }
    }
    m[i] = '\0';
    return m[0];
}

int liremot(char **mot, int *size)
{
    int c;

    do {
       c = readword(mot, size);
    } while (OpenMacro(*mot));
    return c;
}

/* skip C comments */
int negligerComment(int c)
{
    if (c == '/') {
         if ((c = lirecar()) == '*') {
	     for (c = lirecar(); ; ) {
	        while (c != '*') {
                   if (c == EOF) {
                      fatalerror("unclosed comment");
		   } else if (c == '\n')
		      nlignes++;
		   c = lirecar();
		}
		if ((c = lirecar()) == '/')
		   break;
             } 
	     return 1;
          } else
             remettrecar(c);
       }
       return 0;
}

/* Backup a quote, this feature is used for macro definition ... */
#define MAXQUOTE 90
static char quote[MAXQUOTE];

int negligerGuillemets(int c)
{
    register char *ptq = quote;
    int TooLargeQuote = 0;
    /* case of a _true_ quote "something ..." */
    if (c == '"') {
       *ptq++ = '"';
       while ((c = lirecar()) != '"') {
	  if (c == EOF) {
	     fatalerror("unclosed quote, line %d\n",nlignes);
	  } else if (c == '\n')
	     nlignes++;
	  if (ptq <= quote + MAXQUOTE - 2)
	     *ptq++ = c;
	  else
	     TooLargeQuote = 1;
       }
       *ptq++ = '"';
       *ptq = '\0';
       if (TooLargeQuote)
	  fprintf(stderr, "warning : omitting end of a too lage quote\n");
       return 1;
    /* case of the '"' character ... */
    } else if (c == '\'') {
       while ((c = lirecar()) != '\'' && c != EOF)
	  ;
       if (c == EOF)
	  fatalerror(" guillemets non fermees, ligne %d\n",nlignes);
       while ((c = lirecar()) == '\'')
	  ;
       remettrecar(c);
       return 1;
    }
    return 0;
}

char *getQuote(void)
{
    return quote;
}

#undef MAXQUOTE

static char *tamp;
static int ptamp = 0;
FILE *entry;

int lirecar(void)
{
     return (ptamp > 0) ? tamp[--ptamp] : getc(entry);     
}

int lirewspace()
{
     register int c;
     while (isspace(c = lirecar()))
        if (c == '\n')
	   nlignes++;
     return c;
}


/* remet un caractere sur l'entree */
void remettrecar(int c)
{
     static size = 0;
     const step = 50;
     if (ptamp >= size)
          if ((tamp = (char *) realloc ((void *) tamp, (size += step) *
					sizeof(char))) == NULL)
	       fatalerror("remettrecar, trop de caracteres\n");
     tamp[ptamp++] = c;
}

/* put s back to he input */
void remettrech(char *s)
{
     if (*s)
        remettrech(s+1);
     if (*s)
        remettrecar(*s);
}

static void setInput(FILE *file)
{
      entry = file;
}

list *li = NULL;

/* structure to store the description of each structure requested by
   members of. */
typedef struct {
     Variable *var;
     char *name;
     char *func;
     int line;
} MemberInf;

static void storeCopy(Variable *v, char *name, char *func)
{
     MemberInf *tmp;
     tmp = (MemberInf *) pushG(sizeof(MemberInf), &li);
     tmp->var = v;
     tmp->name = dupchaine(name);
     tmp->func = dupchaine(func);
     tmp->line = nlignes;
}

static void affDefCopy(void)
{
     MemberInf *dat;

     dat = (MemberInf *) getG(li);
     if (dat != NULL) {
        printf("\n\n"
	   "\t/********************************************************\n"
	   "\t *        Statistics about members of a structure.      *\n"
           "\t ********************************************************/\n"
	   "\n");
     }
     while (dat != NULL) {
        putchar('\n');
        printf(" the members of %s defined in %s (line %d of %s)\n", 
	       dat->name, dat->func, dat->line, dat->var->var.file);
	printf("are used with these frequencies\n");
        affDefMembers(dat->var);
	printf(
	" end of the statistics about the members of %s in %s\n", 
	dat->name,  dat->var->var.file);
	/* getting the next element ... */
	popG(&li);
	dat = (MemberInf *) getG(li);
     }
}

/* This is the help information */
static void printHelp(void)
{
     printf(
     "This is a software called cvi to provide information about different\n"
     "variables and data types in ANSI/ISO C programs.\n"
     "Copyright (c) 1999 Chepelianskii Alexei\n"
     "alexei@pttlx6.ups-tlse.fr\n"
     "This program is under the GNU General Public License.\n\n"
     "The syntax of the options that can be used to obtain useful\n"
     "information has the form:\n"
     "  -o namelist  -O, o to toggle it,\n"
     "where 'o' represents one of the following ASCII characters, and 'O'\n"
     "its upper case\n"
     "  n : to select a variable name,\n"
     "  s : to require information about a structure/union,\n"
     "  e : to treat C enumerations,\n"
     "  p : for C preprocessor constants,\n"
     "  t : to provide information about typedefs\n"
     "  m : to explain how structure members are used;\n"
     "      to apply -m recursively to structure members use the syntax:\n"
     "      StrName/MemberToParse\n"
     "All these options allow -b to specify a function where the information\n"
     "should be looked for and -n to suppress that condition.\n"
     "  b : outside of any bloc specify a list of fuctions to treat,\n"
     "  f : specify a list of files to treat.\n\n"
     "Here is an example: cvi e -f test.c -F -n -b func var -N\n"
     "to obtain information about how 'var' is used in the function 'func'\n"
     "of the program 'test.c', e asks the program to print enumeration\n"
     "definitions, this feature is disabled by default.\n"
     "Moreover -min and -max options are available to find variables that\n"
     "are very used or not at all.\n\n"
     "Hopping this software will contribute to make programming\n"
     "easier for you.\n"
     "\tThe author.\n\n"
     );
     exit(0);
}


